#![allow(dead_code)]

use serde::{Deserialize, Serialize};
use specs::{
    error::NoError,
    hibitset::BitSet,
    saveload::{ConvertSaveload, Marker},
    Component, Entity, HashMapStorage, VecStorage,
};

use crate::{components::ShipCount, misc::FlaggedStorage};

#[derive(Default, Debug, Clone)]
pub struct Fleet {
    owned: BitSet,
    count: ShipCount,
}

#[derive(Copy, Clone, Debug)]
pub struct Owned {
    owner: Entity,
}

#[derive(Serialize, Deserialize)]
pub struct OwnedData<M> {
    pub owner: M,
}

impl Fleet {
    #[inline]
    pub fn owned(&self) -> &BitSet {
        &self.owned
    }

    #[inline]
    pub(crate) fn owned_mut(&mut self) -> &mut BitSet {
        &mut self.owned
    }

    #[inline]
    pub fn count(&self) -> &ShipCount {
        &self.count
    }

    #[inline]
    pub(crate) fn count_mut(&mut self) -> &mut ShipCount {
        &mut self.count
    }
}

impl Component for Fleet {
    type Storage = HashMapStorage<Self>;
}

impl Owned {
    pub fn new(owner: Entity) -> Self {
        Self { owner }
    }

    pub fn owner(&self) -> Entity {
        self.owner
    }

    pub(crate) fn set_owner(&mut self, owner: Entity) {
        self.owner = owner;
    }
}

impl Component for Owned {
    type Storage = FlaggedStorage<Self, VecStorage<Self>>;
}

impl<M> ConvertSaveload<M> for Owned
where
    for<'de> M: Marker + Serialize + Deserialize<'de>,
{
    type Data = OwnedData<M>;
    type Error = NoError;

    fn convert_into<F>(&self, mut ids: F) -> Result<Self::Data, Self::Error>
    where
        F: FnMut(Entity) -> Option<M>,
    {
        let owner = ids(self.owner).unwrap();

        Ok(OwnedData { owner })
    }

    fn convert_from<F>(data: Self::Data, mut ids: F) -> Result<Self, Self::Error>
    where
        F: FnMut(M) -> Option<Entity>,
    {
        let owner = ids(data.owner).unwrap();

        Ok(Owned { owner })
    }
}
